# Jurassic Race, a 3d minigame.
# Copyright (C) 2009  Travis Burtrum

# The goal of this game is to knock down as many yetti's as possible
# while not getting eaten by the T-rex.  The roadblocks damage your
# jeep by making it not accelerate as fast and therefore increase
# the chance that you will be eaten.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from pandac.PandaModules import loadPrcFileData 
# Uncomment this if you want fullscreen, it works!
#loadPrcFileData("", "fullscreen 1")
loadPrcFileData("", "window-title Jurassic Race")

import direct.directbase.DirectStart
from pandac.PandaModules import *
from direct.interval.IntervalGlobal import *
from direct.gui.OnscreenText import OnscreenText
from direct.actor.Actor import Actor
from direct.task.Task import Task
from direct.showbase.DirectObject import DirectObject
import random, sys, os, math, cPickle

START_LEVEL = 0
NUM_LEVELS = 3

MAX_SPEED = 100
ACCEL = 1500
DECEL_BRAKE = ACCEL*3
DECEL_COAST = ACCEL*0.75
# the following is in addition to whatever other acceleration is being applied
DECEL_TURN = ACCEL*0.75
TURN_RATE = 100
CAM_SPEED = 50
TREX_SPEED = 15

ROADBLOCK_SLOWDOWN = 0.5
ROADBLOCK_ACCEL_DAMAGE = 0.85
TARGET_SCORE = 50
TARGET_MOVING_SCORE = TARGET_SCORE + (TARGET_SCORE*0.5)

TRACK_SCALE = 5
JEEP_FAXLEY = -6.7
JEEP_AXLE_DIST = 11.4  # distance between the two axles on the jeep

FOG_COLOR = ((165/255.0), (180/255.0), (223/255.0))
FOG_DENSITY = 0.0025

LAMP_SCALE = 0.04

#modP = Point3(0,0,0.4)
#modH = -90
FLAG_SCALE = 0.01
FLAG_P = [( VBase3(23.6152, -0.692343, 5.14044) , 74.4902496338 ),
          ( VBase3(27.6114, -1.55654, 5.13864) , 76.5535583496 )]

ROADBLOCK_SCALE = 0.04
ROADBLOCK_RMOD = 1.5
ROADBLOCK_BRMOD = 30
ROADBLOCK_DROP = Point3(0, 0, (7*ROADBLOCK_SCALE))
ROADBLOCK_TIME = 0.5

# modP = Point3(0,0,0.2)
TARGET_SCALE = 0.04
TARGET_RMOD = 1.1
TARGET_BRMOD = 30
TARGET_DROP = Point3(0, 0, (10*TARGET_SCALE))
TARGET_TIME = 0.5
TARGET_MOVE = 4

JEEP_STARTPOS = Point3(25.6224, -1.974, 4.64562)*TRACK_SCALE
JEEP_STARTH = 168
TREX_STARTPOS = Point3(25.6224, -10, 4.64562)*TRACK_SCALE
CAM_MOD = Point3(-2.42535, -9.70142, 2)
START_MOD = Point3(-50,-200,70)

# Function to put instructions on the screen.
def addHudLine(pos, msg, scale=0.06):
    return OnscreenText(text=msg, style=3, fg=(1,0,0,1),
			pos=(1.3, 0.95 - (pos * 0.07) ), align=TextNode.ARight, scale = scale, mayChange=True)

# Function to put title on the screen.
def addTitle(text):
    return OnscreenText(text=text, style=1, fg=(1,1,1,1),
	                pos=(1.3,-0.95), align=TextNode.ARight, scale = .07)
    
def getCollisionEntryZold(colHandlQ, name = "Curve"):
    entries = []
    for i in range(colHandlQ.getNumEntries()):
        entry = colHandlQ.getEntry(i)
        entries.append(entry)
    entries.sort(lambda x,y: cmp(y.getSurfacePoint(render).getZ(),
                                 x.getSurfacePoint(render).getZ()))
    if (len(entries)>0) and (entries[0].getIntoNode().getName() == name):
        return entries[0].getSurfacePoint(render).getZ()
    else:
        return None
    
def getCollisionEntryZ(colHandlQ, name = "Curve"):
    for i in range(colHandlQ.getNumEntries()):
        entry = colHandlQ.getEntry(i)
        if entry.getIntoNode().getName() == name:
            return entry.getSurfacePoint(render).getZ()
    return None
    
def moveForSpeed(dt, speed, object):
    backward = object.getNetTransform().getMat().getRow3(1)
    backward.setZ(0)
    backward.normalize()  
    object.setPos(object.getPos() - backward*(dt*speed))
    
def setupCollisionHandler(object, name='un-named CollisionNode', origin=Point3(0,0,10000)):
    colRay = CollisionRay()
    colRay.setOrigin(origin)
    colRay.setDirection(0,0,-1)
    colNode = CollisionNode(name)
    colNode.addSolid(colRay)
    colNode.setFromCollideMask(BitMask32.bit(0))
    colNode.setIntoCollideMask(BitMask32.allOff())
    colNodeNp = object.attachNewNode(colNode)
    colHandler = CollisionHandlerQueue()
    base.cTrav.addCollider(colNodeNp, colHandler)
    return colHandler

def placeModel(name, pos, h=0, parent=render, scale=1):
    model = loader.loadModel(name)
    model.setScale(scale)      
    model.reparentTo(parent)
    model.setPos(pos)
    model.setH(h)
    return model

def tup2Pnt3(tuple):
    return Point3(tuple[0], tuple[1], tuple[2])

def placeModels(name, array, parent, scale, drop=Point3(0,0,0) ):
    ret = []
    for entry in array:
        ret.append(placeModel(name, (tup2Pnt3(entry[0]) + drop), entry[1], parent, scale))
    return ret    

def setupCollider(object, nodeName, intoMask, fromMask=None, pMod=Point3(0,0,0), rMod=1, inPattern=None, outPattern=None, sphere=True):
    
    bounds = object.getChild(0).getBounds()
    center = bounds.getCenter()+pMod
    radius = bounds.getRadius()*rMod
    
    cNode = CollisionNode(nodeName)
    if sphere:
        solid = CollisionSphere(Point3(center),radius)
    else:
        aPos = Point3(center.getX(), center.getY() - radius, center.getZ())
        bPos = Point3(center.getX(), center.getY() + radius, center.getZ())
        #solid = CollisionSegment(aPos.getX(), ay, az, bx, by, bz)
        solid = CollisionSegment(aPos, bPos)
    cNode.addSolid(solid)
    if fromMask != None:
        cNode.setFromCollideMask(fromMask)
    if intoMask != None:
        cNode.setIntoCollideMask(intoMask)
    cNodePath = object.attachNewNode(cNode)
    
    # uncomment this to show the collision objects
    #cNodePath.show()
    
    if inPattern != None or outPattern != None:
        cHandler = CollisionHandlerEvent()
        if inPattern != None:
            cHandler.addInPattern(inPattern)
        if outPattern != None:
            cHandler.addOutPattern(outPattern)
        base.cTrav.addCollider(cNodePath, cHandler)
        return cHandler

class World(DirectObject):

    def __init__(self):
        self.roadblocks = []
        self.targets = []
        self.points = 0
        self.level = START_LEVEL
        self.loadLevel(self.level)
        self.loadModels()
        self.setupCollisions()

        helpString = "ESC: Quit\n[Left Arrow]: Turn Left\n[Right Arrow]: Turn Right\
        \n[Up Arrow]: Accelerate\n[Down Arrow]: Brake/Reverse\n[f]: Toggle Fog\n[h]: This Text"
        
        self.helpText = OnscreenText(text=helpString, pos=(-1.3, .95), fg=(0, 0, 1, 1),
                                            align=TextNode.ALeft, scale=.05)
        self.helpText.hide()
        # setup keys        
        self.accept("escape", sys.exit)
        self.accept("h", self.toggleHelp)   
        self.accept("i", self.reset)     
        self.accept("f", self.toggleFog)
        
        self.keyMap = {"left":0, "right":0, "accel":0, "brake":0, "cam-left":0, "cam-right":0, "cam-down":0, "cam-up":0}
        self.setupKey("arrow_left", "left")
        self.setupKey("arrow_right", "right")
        self.setupKey("arrow_up", "accel")
        self.setupKey("arrow_down", "brake")
        
        self.setupKey("a", "cam-left")
        self.setupKey("e", "cam-right")
        self.setupKey(",", "cam-up")
        self.setupKey("o", "cam-down")
        
        # the following is for testing, set TREX_SPEED to != 0 for production
        if TREX_SPEED == 0:
            self.enableMapping()

        self.runGame = False
        taskMgr.add(self.move,"moveTask")
        taskMgr.add(self.lookTask,"lookTask")
        
        campos = base.camera.posInterval(5, JEEP_STARTPOS + CAM_MOD, startPos=(JEEP_STARTPOS + START_MOD) )
        startint = Sequence(campos, Func(self.startGame, True))
        startint.start()
        
    def startGame(self, firstRun=False):
        self.runGame = True
        if firstRun:
            self.timeInt = Sequence(Wait(1), Func(self.incTime))
            self.timeInt.loop()
        
    def incTime(self):
        self.time += 1
        self.timeTxt.setText('Time: '+`int(self.time / 60)`+':'+`int(self.time % 60)`)
        
    def toggleHelp(self):
        if self.helpText.isHidden():
            self.helpText.show()
        else:
            self.helpText.hide() 
            
    def toggleFog(self):
        if self.fogOn:
            render.clearFog()
            self.fogOn = False
        else:
            render.setFog(self.fog)  
            self.fogOn = True
       
    def loadLevel(self, num):
        fname = './levels/'+`num`+'.lvl'
        try:
            fileHandle = open(fname, 'r')
        except:
            print fname, 'is not a file, loading blank configuration'
            self.objects = {'target':[],'roadblock':[]}
            return
        self.objects = {}
        self.objects = cPickle.load(fileHandle)
        fileHandle.close()

    def nextLevel(self, reset=False):
        self.runGame = False
        self.speedo.setText('0 MPH')
        self.jeep.stop()
        self.trex.stop()
        camup = base.camera.posInterval(5, JEEP_STARTPOS + Point3(0,0,100) )
        campos = base.camera.posInterval(5, JEEP_STARTPOS + CAM_MOD)
        if reset:
            self.level = 0
            self.points = 0
            self.resetTxt.setText('You lose, go back to first level!')
        else:
            self.level += 1
            self.level %= NUM_LEVELS
            self.resetTxt.setText('You win, advancing to level '+`self.level+1`+'!')
        self.resetTxt.show()
        int = Sequence(camup, campos, Func(self.finishNextLevel))
        int.start()
        
    def finishNextLevel(self):
        self.resetTxt.hide()
        self.loadLevel(self.level)
        self.reset()
           
    def resetVars(self):
        # Game state variables
        self.prevtime = 0
        self.modelInterval = {}
        self.isMoving = False 
        self.jeepSpeed = 0  
        self.trexSpeed = 0
        self.jeepAccel = ACCEL
        self.time = 0
        self.targetsKilled = 0
        
        self.levelTxt.setText('Level: '+`self.level+1`)
        self.timeTxt.setText('Time: 0:00')
        self.completeTxt.setText('Complete: 0%')
        self.pointsTxt.setText('Points: '+`int(self.points)`)
        self.damageTxt.setText('Damage: 0%')
        
    def reset(self, firstRun=False):
        self.resetVars()        
        
        for obj in self.roadblocks:
            obj.remove()
            
        for obj in self.targets:
            obj.remove()
        
        self.roadblocks = placeModels('models/roadblock/roadblock', self.objects['roadblock'], self.racetrack, ROADBLOCK_SCALE, -ROADBLOCK_DROP)
        self.targets = placeModels('models/yeti/yeti', self.objects['target'], self.racetrack, TARGET_SCALE, -TARGET_DROP)
        print 'level', self.level
        print 'loaded', len(self.roadblocks), 'roadblocks'
        print 'loaded', len(self.targets), 'targets'
        
        self.jeep.setPos(JEEP_STARTPOS)
        self.jeep.setH(JEEP_STARTH)
        self.floater.setPos(self.jeep.getPos())
        self.floater.setZ(self.jeep.getZ() + 2.0)
        self.trex.setPos(TREX_STARTPOS)
        self.trex.stop()
        self.trex.lookAt(self.jeep)
        # it is backwards, flip it around
        self.trex.setH((self.trex.getH() + 180) % 360)
        
        base.camera.setPos(JEEP_STARTPOS + CAM_MOD)
        
        if not firstRun:
            self.resetCollisions()
            self.runGame = True
            
        
    def resetCollisions(self):
        for obj in self.roadblocks:
            setupCollider(obj, "roadblock", BitMask32(0x03), rMod=ROADBLOCK_RMOD)
            setupCollider(obj, "roadblock-up", BitMask32(0x03), rMod=ROADBLOCK_BRMOD)
            
        for obj in self.targets:
            setupCollider(obj, "target", BitMask32(0x03), rMod=TARGET_RMOD)
            setupCollider(obj, "target-up", BitMask32(0x03), rMod=TARGET_BRMOD)
       
    def loadModels(self):
        base.camLens.setFar(base.camLens.getFar()*TRACK_SCALE)
        
        # load racetrack
        self.racetrack = loader.loadModel("models/env/racetrack")
        self.racetrack.setScale(TRACK_SCALE)      
        self.racetrack.reparentTo(render)
        self.racetrack.setPos(0,0,0)
        
        # load the sky
        self.sky = loader.loadModel("models/env/sky")
        self.sky.setScale(180)      
        self.sky.reparentTo(self.racetrack)
        self.sky.setPos(0,0,0)
        
        # Create the jeep
        self.jeep = Actor("models/jeep/jeep", {"drive":"models/jeep/jeep-loop", "start":"models/jeep/jeep-start2"})
        self.jeep.reparentTo(render)
        self.jeep.setScale(.2)
        
        self.trex = Actor("models/trex/trex", {"run":"models/trex/trex-run", "eat":"models/trex/trex-eat"})
        self.trex.reparentTo(render)
        self.trex.setScale(.2)

        # Create a floater object to aim the camera at        
        self.floater = NodePath(PandaNode("floater"))
        
        # Set up the camera        
        base.disableMouse()
        
        self.resetTxt = OnscreenText(text="Blank", pos=(0, 0), fg=(0,0,1,1),
                                            align=TextNode.ACenter, scale=0.17, mayChange=True)
        self.resetTxt.hide()
        self.speedo = OnscreenText(text='0 MPH', pos=(0, 0.88), fg=(1,1,1,1),
                                            align=TextNode.ACenter, scale=0.17, mayChange=True)
        self.levelTxt = addHudLine(0, 'Level: 1')
        self.timeTxt = addHudLine(1, 'Time: 00:00')
        self.completeTxt = addHudLine(2, 'Complete: 0%')
        self.pointsTxt = addHudLine(3, 'Points: 0')
        self.damageTxt = addHudLine(4, 'Damage: 0%')
                                            
        self.fog = Fog("Scene-wide exponential Fog")
        self.fog.setColor(*FOG_COLOR)
        self.fog.setExpDensity(FOG_DENSITY)
        base.setBackgroundColor(*FOG_COLOR)
        render.setFog(self.fog)
        self.fogOn = True
        #self.fogOn = False
        
        # now load objects into the map
        self.flags = placeModels('models/flag/flag', FLAG_P, self.racetrack, FLAG_SCALE)
        
        # first run
        self.reset(True)
        
    def setupCollisions(self):
        # create CollisionTraverser
        self.cTrav = CollisionTraverser()
        # make it the default
        base.cTrav = self.cTrav

        self.jeepFaxelHandler = setupCollisionHandler(self.jeep, 'jeepFrontaxle', Point3(0,JEEP_FAXLEY,10000))
        self.jeepRaxelHandler = setupCollisionHandler(self.jeep, 'jeepRearaxle', Point3(0,JEEP_FAXLEY+JEEP_AXLE_DIST,10000))
        self.jeepGroundHandler = setupCollisionHandler(self.jeep, 'jeepRay')
        self.trexGroundHandler = setupCollisionHandler(self.trex, 'trexRay')
        self.camGroundHandler = setupCollisionHandler(base.camera, 'camRay')
        
        # setupCollider(object, nodeName, intoMask, fromMask=None, pMod=Point3(0,0,0), rMod=1, inPattern=None, outPattern=None )
        self.trexCHandler = setupCollider(self.trex, "trex", BitMask32(0x00), BitMask32(0x02), Point3(0,-13,8), 0.15, "ate-%in" )
        # for collisions with trex
        self.jeepSphereCHandler = setupCollider(self.jeep, "jeep", BitMask32(0x02), BitMask32.allOff(), Point3(0,0,-5))
        # for collisions with objects
        self.jeepSegmentCHandler = setupCollider(self.jeep, "jeepSegment", BitMask32.allOff(), BitMask32(0x03), Point3(0,0,-5), 1, "hit-%in", "out-%in", False )
        
        self.resetCollisions()
        
        self.accept("ate-jeep", self.eatJeep)
        self.accept("hit-roadblock", self.jeepHitRoadblock)
        self.accept("hit-roadblock-up", self.roadblockUp)
        self.accept("out-roadblock-up", self.roadblockDown)
        self.accept("hit-target", self.jeepHitTarget)
        self.accept("hit-target-up", self.targetUp)
        self.accept("out-target-up", self.targetDown)
        
    def moveModel(self, model, time, startPos, drop, mpos=None, mtime=TARGET_MOVE):
        pos = startPos + drop
        
        # dictionary for all running intervals, model is the key
        if model in self.modelInterval:
            entry = self.modelInterval[model]
            entry[0].finish()
            mint = entry[1]
            if mint != None:
                # we were looping, now we need to delete it ourself
                del self.modelInterval[model]
                mint.finish()
        int = model.posInterval(time, pos)
        mint = None
        if mpos != None:
            to = model.posInterval(mtime, mpos)
            back = model.posInterval(mtime, pos)
            mint = Sequence(to, back)
            int = Sequence(int, Func(mint.loop))
        else:
            # if we aren't looping, let the sequence delete it
            int = Sequence(int, Func(self.delModelInterval, model))
        self.modelInterval[model] = (int, mint)
        # start int
        self.modelInterval[model][0].start()
        
    def delModelInterval(self, model):
        del self.modelInterval[model]
        
    def eatJeep(self,cEntry):
        self.trex.stop()
        self.trex.play('eat')
        self.trexSpeed = 0
        # reset the game
        self.nextLevel(True)
        
    def jeepHitTarget(self,cEntry):
        target = cEntry.getIntoNodePath().getParent()
        entry = self.objects['target'][self.targets.index(target)]
        #self.targets.remove(target) # no real reason to remove from list, keeps index integrity
        target.remove()
        # if it is 3, then its a moving target
        if len(entry) == 3:
            self.points += TARGET_MOVING_SCORE
        else:
            self.points += TARGET_SCORE
        self.pointsTxt.setText('Points: '+`int(self.points)`)
        self.targetsKilled += 1
        # 1.0 to make this a floating point calculation
        complete = int( (1.0*self.targetsKilled / len(self.targets))*100 )
        self.completeTxt.setText('Complete: '+`complete`+'%')
        if complete == 100:
            self.nextLevel()
            
        
    def targetUp(self,cEntry):
        target = cEntry.getIntoNodePath().getParent()
        if target not in self.targets:
            return
        entry = self.objects['target'][self.targets.index(target)]
        mpos = None
        if len(entry) == 3:
            mpos = tup2Pnt3(entry[2])
        ipos = tup2Pnt3(entry[0])
        interval = self.moveModel(target, TARGET_TIME, ipos-TARGET_DROP, TARGET_DROP, mpos)
    
    def targetDown(self,cEntry):
        target = cEntry.getIntoNodePath().getParent()
        if target not in self.targets:
            return
        ipos = tup2Pnt3(self.objects['target'][self.targets.index(target)][0])
        interval = self.moveModel(target, TARGET_TIME, ipos, -TARGET_DROP)
        
    def jeepHitRoadblock(self,cEntry):
        roadblock = cEntry.getIntoNodePath().getParent()
        #self.roadblocks.remove(roadblock) # no real reason to remove from list, keeps index integrity
        roadblock.remove()
        self.jeepSpeed *= ROADBLOCK_SLOWDOWN
        self.jeepAccel *= ROADBLOCK_ACCEL_DAMAGE
        # 1.0 to make this a floating point calculation
        damage = ( (ACCEL - self.jeepAccel) / ACCEL)*100
        self.damageTxt.setText('Damage: '+`int(damage)`+'%')
        
    def roadblockUp(self,cEntry):
        roadblock = cEntry.getIntoNodePath().getParent()
        if roadblock not in self.roadblocks:
            return
        ipos = tup2Pnt3(self.objects['roadblock'][self.roadblocks.index(roadblock)][0])
        interval = self.moveModel(roadblock, ROADBLOCK_TIME, ipos-ROADBLOCK_DROP, ROADBLOCK_DROP)
    
    def roadblockDown(self,cEntry):
        roadblock = cEntry.getIntoNodePath().getParent()
        if roadblock not in self.roadblocks:
            return
        ipos = tup2Pnt3(self.objects['roadblock'][self.roadblocks.index(roadblock)][0])
        interval = self.moveModel(roadblock, ROADBLOCK_TIME, ipos, -ROADBLOCK_DROP)
        
    def setupKey(self, key, value):
        self.accept(key, self.keyMap.update, [{value : 1}])
        self.accept(key+"-up", self.keyMap.update, [{value : 0}])
    
    def speedForAccel(self, dt, accel):
        self.jeepSpeed = self.jeepSpeed + (0.5 * accel * (dt**2) )

    def lookTask(self, task):
        base.camera.lookAt(self.floater)
        return Task.cont
    # Accepts arrow keys to move either the player or the menu cursor,
    # Also deals with grid checking and collision detection
    def move(self, task):

        dt = task.time - self.prevtime
        self.prevtime = task.time
        
        if not self.runGame:
            return Task.cont

        # If the camera-left key is pressed, move camera left.
        # If the camera-right key is pressed, move camera right.

        base.camera.lookAt(self.jeep)
        camright = base.camera.getNetTransform().getMat().getRow3(0)
        camright.normalize()
        if (self.keyMap["cam-left"]!=0):
            base.camera.setPos(base.camera.getPos() - camright*(dt*CAM_SPEED))
        if (self.keyMap["cam-right"]!=0):
            base.camera.setPos(base.camera.getPos() + camright*(dt*CAM_SPEED))
        camright = base.camera.getNetTransform().getMat().getRow3(2)
        camright.normalize()
        if (self.keyMap["cam-down"]!=0):
            base.camera.setPos(base.camera.getPos() - camright*(dt*CAM_SPEED))
        if (self.keyMap["cam-up"]!=0):
            base.camera.setPos(base.camera.getPos() + camright*(dt*CAM_SPEED))

        # save jeep's initial position so that we can restore it,
        # if it's no longer on the ground

        startpos = self.jeep.getPos()

        # If a move-key is pressed, move jeep in the specified direction.

        if self.keyMap["left"]:# and self.isMoving:
            self.jeep.setH((self.jeep.getH() + dt*TURN_RATE) % 360)
            if self.jeepSpeed > 1:
                self.speedForAccel(dt, -DECEL_TURN)
            elif self.jeepSpeed < -1:
                self.speedForAccel(dt, DECEL_TURN)
        elif self.keyMap["right"]:# and self.isMoving:
            self.jeep.setH((self.jeep.getH() - dt*TURN_RATE) % 360)
            if self.jeepSpeed > 1:
                self.speedForAccel(dt, -DECEL_TURN)
            elif self.jeepSpeed < -1:
                self.speedForAccel(dt, DECEL_TURN)
        # accelerating
        if self.keyMap["accel"]:
            # if we are actually accelerating, speed will be greater than or 0
            if self.jeepSpeed >= 0:
                self.speedForAccel(dt, self.jeepAccel)
                if self.jeepSpeed > MAX_SPEED:
                    self.jeepSpeed = MAX_SPEED
            # if the speed is less than 0, then we are actually braking, use DECEL_BRAKE    
            else:
                self.speedForAccel(dt, DECEL_BRAKE)
        elif self.keyMap["brake"]:
            # if we are actually braking, the speed will be greater than 0
            if self.jeepSpeed > 0:
                self.speedForAccel(dt, -DECEL_BRAKE)
            # if the speed is less than or 0, then we are actually accelerating backwards, use ACCEL    
            else:
                self.speedForAccel(dt, -self.jeepAccel)
                if self.jeepSpeed < -MAX_SPEED:
                    self.jeepSpeed = -MAX_SPEED
        # coasting
        else:
            if self.jeepSpeed > 0:
                self.speedForAccel(dt, -DECEL_COAST)
            elif self.jeepSpeed < 0:
                self.speedForAccel(dt, DECEL_COAST)
            # if the speed is very close to 0, set it to 0
            if self.jeepSpeed < 0.1 and self.jeepSpeed > -0.1:
                self.jeepSpeed = 0
                self.speedo.setText('0 MPH')
        
        # only update the position if the speed is greater than 0
        if self.jeepSpeed != 0:
            # this starts the trex only after we start, it's fair that way
            if self.trexSpeed == 0:
                self.trex.loop("run")
                self.trexSpeed = TREX_SPEED
            if not self.isMoving:
                self.jeep.loop("drive")
                self.isMoving = True
            #print 'self.jeepSpeed', self.jeepSpeed
            self.speedo.setText(`int(math.fabs(self.jeepSpeed))`+' MPH')
            moveForSpeed(dt, self.jeepSpeed, self.jeep)
        # else the jeep is not moving, stop the animation
        else:
            if self.isMoving:
                self.jeep.stop()
                self.isMoving = False

        # If the camera is too far from jeep, move it closer.
        # If the camera is too close to jeep, move it farther.

        camvec = self.jeep.getPos() - base.camera.getPos()
        camvec.setZ(0)
        camdist = camvec.length()
        camvec.normalize()
        if (camdist > 10.0):
            base.camera.setPos(base.camera.getPos() + camvec*(camdist-10))
            camdist = 10.0
        if (camdist < 5.0):
            base.camera.setPos(base.camera.getPos() - camvec*(5-camdist))
            camdist = 5.0            
            
        # aim the t-rex at the jeep
        self.trex.lookAt(self.jeep)
        # it is backwards, flip it around
        self.trex.setH((self.trex.getH() + 180) % 360)
        # save position
        trexPos = self.trex.getPos()
        moveForSpeed(dt, self.trexSpeed, self.trex)

        # Now check for collisions.

        self.cTrav.traverse(render)
        
        # keep trex where it should be
        z = getCollisionEntryZ(self.trexGroundHandler)
        if z != None:
            self.trex.setZ(z)
        else:
            self.trex.setPos(trexPos)

        # keep the jeep where it should be
        z = getCollisionEntryZ(self.jeepGroundHandler)
        if z != None:
            self.jeep.setZ(z)
        else:
            # when you run into an unmoveable object, you stop
            self.jeepSpeed = 0
            self.jeep.setPos(startpos)

        # Keep the camera at one foot above the terrain,
        # or two feet above jeep, whichever is greater.
        z = getCollisionEntryZ(self.camGroundHandler)
        if z != None:
            base.camera.setZ(z + 1.0)
        if (base.camera.getZ() < self.jeep.getZ() + 2.0):
            base.camera.setZ(self.jeep.getZ() + 2.0)
            
        # now get the front and rear axle distances to the ground
        # in order to set the proper P angle
        faxleZ = getCollisionEntryZ(self.jeepFaxelHandler)
        raxleZ = getCollisionEntryZ(self.jeepRaxelHandler)
        # if they are both valid, then calculate and set the new P
        if faxleZ != None and raxleZ != None:
            # adjacent is JEEP_AXLE_DIST, which is never 0
            opp = raxleZ - faxleZ
            p = math.degrees( math.atan( opp / JEEP_AXLE_DIST ) )*TRACK_SCALE
            self.jeep.setP(p)
            
                
        # The camera should look in jeep's direction,
        # but it should also try to stay horizontal, so look at
        # a floater which hovers above jeep's head.
        
        self.floater.setPos(self.jeep.getPos())
        self.floater.setZ(self.jeep.getZ() + 2.0)

        # continue
        return Task.cont
    
    def enableMapping(self):
        self.temp = None
        self.tname = None
        self.tMod = Point3(0,0,0)
        self.accept("y", self.p, ['models/yeti/yeti', TARGET_SCALE, 'target', Point3(0,0,0.2)])
        self.accept("r", self.p, ['models/roadblock/roadblock', ROADBLOCK_SCALE, 'roadblock'])
        self.accept("l", self.p, ['models/lamp/lamp', LAMP_SCALE])
        self.accept("1", self.z, [-1]) # move temp down
        self.accept("2", self.z)       # move temp up
        self.accept("m", self.m)
        self.accept("c", self.commit)
        self.accept("d", self.dumpLevel)
        
    def p(self, name, scale, key=None, modP=Point3(0,0,0), modH=0):
        pos = self.jeep.getPos()/TRACK_SCALE
        h = self.jeep.getH()
        #print 'setPos(', pos, ')', 'setH(', h, ')'
        # def placeModel(name, pos, h=0, parent=render, scale=1):
        self.modP = modP
        self.moveP = None
        newP = pos + modP
        newH = h + modH
        self.tname = key
        print name
        print '(', newP, ',', newH, ')'
        if self.temp != None:
            self.temp.remove()
            self.tMod = Point3(0,0,0)
        self.temp = placeModel(name, newP, newH, self.racetrack, scale)        
        
    def z(self, one=1):
        if self.temp == None:
            return
        scale = self.temp.getScale().getX()
        mod = Point3(0,0,one)*scale
        self.tMod += mod
        print 'tMod', self.tMod
        self.temp.setPos(self.temp.getPos()+mod)
        
    def m(self):
        if self.tname != 'target':
            print 'self.temp not a target'
            return
        pos = self.jeep.getPos()/TRACK_SCALE
        newP = pos+self.modP
        print self.tname
        print '(', newP, ')'
        self.moveP = newP
        
    def commit(self):
        if self.temp == None or self.tname == None:
            print 'nothing to commit'
            return
        pos = self.temp.getPos()
        h = self.temp.getH()
        pos = (pos.getX(), pos.getY(), pos.getZ())
        entry = ( pos, h )
        if self.moveP == None:
            entry = ( pos, h )
        else:
            entry = ( pos, h, (self.moveP.getX(), self.moveP.getY(), self.moveP.getZ()) )       
        print 'entry', entry
        self.objects[self.tname].append(entry)
        # set this to stop it from being removed from the screen
        self.temp = None
        
    def dumpLevel(self):
        print 'dumping level'
        fileHandle = open('levels/fresh.lvl', 'w')
        cPickle.dump(self.objects, fileHandle)
        fileHandle.close()
        print 'succesfully dumped to levels/fresh.lvl'

w = World()
run()
